﻿// ======
// small PixelLiner file (.pxls) Loader
// by cloba.u 2019
// ======
using System.Collections;
using System.Collections.Generic;
using PixelLiner.PixelLinerLib;
using UnityEngine;

namespace PixelLiner {
    public class PxlCharacter {
        public string title = "";
        public static readonly uint file_header1 = 0x7741a8ff;
        public static readonly string file_header2 = "PXLS";
        public string error_str;
        public string external_png_header = "";

        public Dictionary<string, PxlImage> OImg; // PxlImage
        public PxlPose[] APose;

        enum LP : int {
            INIT = 0,
            ERROR = -1,
            PHASE_INIT = 10,
            P_IMG = 11,
            P_POSE = 12,
            P_IMG_PACKED = 15,
            P_IMG_VECTOR = 16,
            RENDER_FRAME = 20,
            RELOAD_IMAGE = 21,
            COMPLETE = 128,
        }

        TextAsset LoadingAsset;
        ByteArray LoadBytes;
        ByteArray BaS;
        PxlsTexture[] AExternalTexture;
        LP load_phase = LP.INIT;
        int load_image_count = -1;
        int load_image_i = 0;
        int load_pose_count = -1;
        public int frame_count = 0;
        public PxlFrame[] AFrame;
        public float pixelsPerUnit = 16.0f;
        public bool autoFlipX = true;
        public bool full_cvs_render = false;
        public bool no_load_external_texture_on_first = false; // 初回ロードの際にテクスチャを読み込みません

        private static readonly int SCT_LEN = 14;

        private const string SEC_IMGS = "%IMGS_SECTION%";
        private const string SEC_IMGD = "%IMGD_SECTION%";
        private const string SEC_IMGV = "%IMGV_SECTION%";
        private const string SEC_POSE = "%POSE_SECTION%";
        private const string SEC_IMGS_PACKED = "%PACK_SECTION%";

        // ======================================================== 
        // MARK: constructor 
        // ========================================================
        public PxlCharacter (string _title) {
            APose = new PxlPose[0];
            OImg = new Dictionary<string, PxlImage> ();
            title = _title;
            external_png_header = PxlsTexture.PxlsTexturePath (title + ".pxls", -1);
            PxlsLoader.assignChara (this);
        }
        // ======================================================== 
        // MARK: method 
        // ========================================================
        public bool loadASync (TextAsset Ta) {
            return loadASync (Ta.bytes);
        }
        public bool loadASync (byte[] bytes) {
            if (!(load_phase == LP.INIT || load_phase == LP.ERROR)) {
                return false;
            }
            /*
            バイト列 Bytes から読み込みを実行します。
            既に読み込まれている場合、または aLoadBytes により既に非同期処理が実行されている場合は false が返ります。
            正常にロードが開始した場合は true が返ります。

            この命令は非同期処理で、この関数実行直後はキャラクターファイルのロードは完了していません。
            その他、何らかのエラーが起こった場合は例外がスローされます。

            */
            LoadBytes = new ByteArray (bytes);
            LoadBytes.position = 0;
            load_phase = LP.INIT;
            BaS = null;
            error_str = "";
            load_image_count = load_pose_count = -1;
            frame_count = 0;
            AFrame = null;
            PxlsLoader.initLoadASync (this);
            return true;
        }
        public void close () {
            // loadASync 終了。
            LoadBytes = null;
            BaS = null;
        }

        public bool isLoading () {
            return load_phase >= LP.INIT && load_phase < LP.COMPLETE;
        }
        public bool isLoadCompleted () {
            return load_phase >= LP.COMPLETE;
        }
        public bool errorOccured () {
            return load_phase == LP.ERROR;
        }
        public int countPoses () {
            return APose.Length;
        }
        public PxlPose getPose (int i) {
            return X.BTW (0, i, APose.Length) ? APose[i] : null;
        }
        public PxlPose getPoseByName (string s) {
            for (int i = APose.Length - 1; i >= 0; i--) {
                if (APose[i].title == s) return APose[i];
            }
            return null;
        }

        public PxlPose[] getPoseArray () {
            return APose;
        }
        public PxlFrame[] ListupFrames () {
            PxlFrame[] _AFrame = new PxlFrame[frame_count];
            int j = 0;
            for (int i = 0; i < APose.Length; i++) {
                j = APose[i].ListupFrames (_AFrame, j);
            }
            return _AFrame;
        }
        public Dictionary<string, PxlImage> getImageObject () {
            return OImg;
        }
        public ByteArray getLoadingBytes () {
            return LoadBytes;
        }
        public void Destroy (bool no_dispose_image = false) {
            if (!no_dispose_image) {
                foreach (KeyValuePair<string, PxlImage> kvp in OImg) {
                    kvp.Value.Destroy ();
                }
                OImg.Clear ();
                releaseLoadedExternalTexture ();
            }
            // foreach (KeyValuePair<string, PxlFrameRenderedImage> kvpr in ORendImg) {
            //     kvpr.Value.Destroy ();
            // }
            loadAbort ();
            close ();
        }

        // ======================================================== 
        // MARK: inner 
        // ========================================================
        bool loadComplete () {
            load_phase = LP.COMPLETE;
            close ();
            return false;
        }
        public bool callError (string txt = "") {
            load_phase = LP.ERROR;
            Debug.Log (txt);
            close ();
            error_str = txt;
            return false;
        }

        internal bool loadProgress (float loadSpeed) {
            if (load_phase == LP.RELOAD_IMAGE) {
                if (ReloadImageProgress ()) {
                    loadComplete ();
                    return true;
                }
                return false;
            }
            if (LoadBytes == null) {
                // 初期化されていない
                callError ("no initialization");
                return true;
            }
            uint len;
            int cnt = (int) (400 * loadSpeed);
            bool reduced = false;
            while (--cnt >= 0 && LoadBytes != null) {
                try {
                    switch (load_phase) {
                        case LP.INIT:
                            if (LoadBytes.readUnsignedInt () != file_header1 || LoadBytes.readMultiByte (4, "utf-8") != file_header2) {
                                callError ("invalid header");
                                return true;
                            }
                            load_phase = LP.PHASE_INIT;
                            cnt = 0;
                            break;
                        case LP.PHASE_INIT:
                            reduced = false;
                            string sec_key = LoadBytes.bytesAvailable > (ulong) SCT_LEN ? LoadBytes.readMultiByte (SCT_LEN, "utf-8") : "";
                            switch (sec_key) {
                                case SEC_IMGD:
                                    callError ("does not support for compressed image data");
                                    return true;
                                case SEC_IMGS:
                                case SEC_IMGS_PACKED:
                                case SEC_IMGV:
                                    load_phase = sec_key == SEC_IMGS_PACKED?LP.P_IMG_PACKED: (sec_key == SEC_IMGV?LP.P_IMG_VECTOR : LP.P_IMG);
                                    // slice by section length
                                    len = LoadBytes.readUnsignedInt ();
                                    BaS = new ByteArray ();
                                    LoadBytes.readBytes (BaS, 0, len);
                                    BaS.position = 0;
                                    load_image_i = 0;
                                    load_image_count = (int) BaS.readUnsignedInt ();

                                    break;
                                case SEC_POSE:
                                    if (load_image_count < 0) {
                                        callError ("pose section must be after image section appear");
                                        return true;
                                    }
                                    load_phase = LP.P_POSE;

                                    // slice by section length
                                    len = LoadBytes.readUnsignedInt (); // !!!
                                    BaS = new ByteArray ();
                                    LoadBytes.readBytes (BaS, 0, len);
                                    BaS.position = 0;
                                    load_pose_count = (int) BaS.readUnsignedInt ();
                                    break;
                                default:
                                    if (load_image_count == 0 && load_pose_count == 0) {
                                        //load_phase = LP.RENDER_FRAME;
                                        loadComplete ();
                                        break;
                                    }
                                    callError ("invalid section header");
                                    return true;
                            }
                            break;
                        case LP.P_IMG_PACKED:
                            if (!reduced) {
                                reduced = true;
                                int mcnt = Mathf.CeilToInt (50 * loadSpeed);
                                if (cnt > mcnt) cnt = mcnt;
                            }
                            if (load_image_count > 0) {
                                load_image_count--;
                                PxlsImgAtlas.readFromBytesStack (BaS, this);
                                load_image_i++;
                            } else {
                                load_phase = LP.PHASE_INIT;
                                PxlsImgAtlas.clearLoadingStack ();
                            }
                            break;
                        case LP.P_IMG:
                            if (!reduced) {
                                reduced = true;
                                int mcnt = Mathf.CeilToInt (50 * loadSpeed);
                                if (cnt > mcnt) cnt = mcnt;
                            }
                            if (load_image_count > 0) {
                                load_image_count--;
                                new PxlImage (this, BaS);
                            } else {
                                load_phase = LP.PHASE_INIT;
                            }
                            break;
                        case LP.P_IMG_VECTOR:
                            if (!reduced) {
                                reduced = true;
                                int mcnt = Mathf.CeilToInt (100 * loadSpeed);
                                if (cnt > mcnt) cnt = mcnt;
                            }
                            if (load_image_count > 0) {
                                load_image_count--;
                                if (!PxlVectorData.progressVectorDataReading (BaS, this)) {
                                    load_phase = LP.PHASE_INIT;
                                }
                            } else {
                                load_phase = LP.PHASE_INIT;
                            }
                            break;
                        case LP.P_POSE:
                            if (!reduced) {
                                reduced = true;
                                // int mcnt = Mathf.CeilToInt (100 * loadSpeed);
                                // if (cnt > mcnt) cnt = mcnt;
                            }
                            if (load_pose_count > 0) {
                                load_pose_count--;
                                PxlPose _Pose = new PxlPose (this);
                                if (_Pose.readFromBytes (BaS)) {
                                    System.Array.Resize (ref APose, APose.Length + 1);
                                    APose[APose.Length - 1] = _Pose;
                                }
                            } else if (load_pose_count == 0) {
                                load_phase = LP.PHASE_INIT;
                            }

                            break;
                            // case LP.RENDER_FRAME:
                            //     if (AFrame == null) {
                            //         AFrame = ListupFrames ();
                            //         load_framerender_count = 0;
                            //     }
                            //     renderFrameImage (AFrame[load_framerender_count]);
                            //     if (++load_framerender_count >= AFrame.Length) {
                            //         // ここで完了
                            //         AFrame = null;
                            //         loadComplete ();
                            //     }
                            //     break;
                        default:
                            break;
                    }
                } catch (System.Exception err) {
                    callError (err.ToString () + " ... bytes reading error occured");
                    return true;
                }
            }
            if (LoadBytes == null) {
                if (!isLoadCompleted ()) {
                    callError ("ByteArray is unexpectedly finished.");
                }
                return true;
            }
            return false;
        }
        public bool ReloadImageProgress () {
            // ロード終了時 true
            if (AExternalTexture == null) return true;
            int ARMX = AExternalTexture.Length;
            for (int i = 0; i < ARMX; i++) {
                var Ptx = AExternalTexture[i];
                if (Ptx.Image == null && Ptx.texture_path != null) {
                    Ptx.assignImage (Resources.Load<Texture2D> (Ptx.texture_path));
                    return false;
                }
            }

            return true;
        }
        internal PxlsTexture loadExternalPngResource (PxlsImgAtlas Atlas, int width, int height, Texture2D TextureEmbed = null) {
            // 外部出力された PNG を返却する。
            PxlsTexture Ptx = null;
            if (AExternalTexture == null) {
                AExternalTexture = new PxlsTexture[1];
            }
            if (AExternalTexture.Length <= load_image_i) {
                System.Array.Resize (ref AExternalTexture, load_image_i + 1);
            }
            if (AExternalTexture.Length > load_image_i) {
                Ptx = AExternalTexture[load_image_i];
                if (Ptx == null) {
                    AExternalTexture[load_image_i] = Ptx = (new PxlsTexture (external_png_header + load_image_i, !no_load_external_texture_on_first));
                }
                Ptx.assignSizeFromBinary (width, height);
            }
            return Ptx;
        }
        public PxlCharacter AddExternalPng (PxlsTexture[] ATx) {
            if (ATx == null) return this;
            int ARMX = ATx.Length,
                pos = 0;
            if (AExternalTexture == null) AExternalTexture = new PxlsTexture[ARMX];
            else {
                pos = AExternalTexture.Length;
                System.Array.Resize (ref AExternalTexture, pos + ARMX);
            }
            for (int i = 0; i < ARMX; i++) {
                AExternalTexture[pos++] = ATx[i];
            }
            return this;

        }
        public PxlsTexture[] getExternalTextureArray () {
            return AExternalTexture;
        }
        public void releaseLoadedExternalTexture () {
            if (AExternalTexture == null) return;
            int ARMX = AExternalTexture.Length;
            for (int i = 0; i < ARMX; i++) {
                var Tx = AExternalTexture[i];
                AExternalTexture[i].Unload (true);
            }
        }

        public bool reloadExternalTexture () {
            // reload がはじまったら true
            if (AExternalTexture == null || isLoading ()) return false;
            int ARMX = AExternalTexture.Length;
            bool ret = false;
            for (int i = 0; i < ARMX; i++) {
                var Tx = AExternalTexture[i];
                if (Tx.Image == null) {
                    ret = true;
                    break;
                }
            }
            if (ret) {
                load_phase = LP.RELOAD_IMAGE;
                PxlsLoader.initLoadASync (this);
            }
            return ret;
        }
        public void assignImage (PxlImage pImg) {
            OImg[pImg.idstr] = pImg;
        }
        public PxlImage getImage (string idstr) {
            if (OImg.ContainsKey (idstr)) {
                return OImg[idstr];
            }
            return null;
        }
        public void loadAbort () {
            //ロード中断。
            if (isLoading ()) {
                callError ("aborted");
            }
        }
    }
}